###
##  Django Proxypay Configurations Methods
#

# django stuff
from django.conf import settings

# proxypay stuff
from proxypay.exceptions import ProxypayKeyError, ProxypayValueError

# =================================================================================================================================================

###
##  Django Proxypay Constant
#

PP_SETTINGS_CONFIG_KEY                  = 'PROXYPAY'
PP_CONFIG_API_TOKEN_KEY                 = 'PRIVATE_KEY'
PP_CONFIG_API_ENTITY_KEY                = 'ENTITY'
PP_CONFIG_API_ENV_KEY                   = 'ENV'
PP_CONFIG_REFS_EXPIRES_DAYS_KEY         = 'REFERENCE_DAYS'
PP_CONFIG_DEFAULT_DEVELOPMENT_ENV       = 'sandbox'
PP_CONFIG_DEFAULT_PRODUCTION_ENV        = 'production'
PP_API_PRODUCTION_BASE_URL              = 'https://api.proxypay.co.ao'
PP_API_DEVELOPMENT_BASE_URL             = 'https://api.sandbox.proxypay.co.ao'
PP_AUTO_PAYMENT_REF_ID                  = 'djpp_reference_id'

# =================================================================================================================================================

def get_configurations():

    """Get and validate configuration for proxypay on settings"""

    if hasattr( settings, PP_SETTINGS_CONFIG_KEY ):
        
        # getting the configuration for proxypay
        configs = eval(f"settings.{PP_SETTINGS_CONFIG_KEY}")
        
        # gettings api environment
        if configs.get(PP_CONFIG_API_ENV_KEY) is None:
            # using env based on settings.DEBUG status
            environment = PP_CONFIG_DEFAULT_DEVELOPMENT_ENV if settings.DEBUG else PP_CONFIG_DEFAULT_PRODUCTION_ENV
        else:
            environment = configs.get(PP_CONFIG_API_ENV_KEY)
            #
            if environment not in [PP_CONFIG_DEFAULT_DEVELOPMENT_ENV, PP_CONFIG_DEFAULT_PRODUCTION_ENV]:
                raise ProxypayValueError(
                    f"{PP_CONFIG_API_ENV_KEY} expected value was {PP_CONFIG_DEFAULT_PRODUCTION_ENV} or {PP_CONFIG_DEFAULT_DEVELOPMENT_ENV}"
                )
        
        # getting api token key
        token = configs.get(PP_CONFIG_API_TOKEN_KEY)
        if not token:
            raise ProxypayValueError(
                f"{PP_CONFIG_API_TOKEN_KEY} is required"
            )

        # getting entity
        entity = configs.get(PP_CONFIG_API_ENTITY_KEY)
        if not entity:
            raise ProxypayValueError(
                f"{PP_CONFIG_API_ENTITY_KEY} is required"
            )
        
        #
        return {
            'environment': environment,
            'entity': entity,
            'token': token,
            'url': PP_API_PRODUCTION_BASE_URL if environment == PP_CONFIG_DEFAULT_PRODUCTION_ENV else PP_API_DEVELOPMENT_BASE_URL
        }

    raise ProxypayKeyError(
        f"{PP_SETTINGS_CONFIG_KEY} key not found on settings"
    )

# =================================================================================================================================================

def get_default_reference_expires_days():

    """  """

    if hasattr( settings, PP_SETTINGS_CONFIG_KEY ):
        
        # getting the configuration for proxypay
        configs = eval(f"settings.{PP_SETTINGS_CONFIG_KEY}")
        
        # 
        if configs.get(PP_CONFIG_REFS_EXPIRES_DAYS_KEY):
            # 
            days = configs.get(PP_CONFIG_REFS_EXPIRES_DAYS_KEY)
            #
            if type(days) is int:
                return days
            #
            if type(days) is str:
                if days.isnumeric():
                    return int(days)
            raise ProxypayValueError(f"{PP_CONFIG_REFS_EXPIRES_DAYS_KEY} value expected to be an integer")
        #
        return False
    #
    raise ProxypayKeyError(
        f"{PP_SETTINGS_CONFIG_KEY} key not found on settings"
    )

# =================================================================================================================================================

def get_private_key():
    #
    if hasattr( settings, PP_SETTINGS_CONFIG_KEY ):
        # getting the configuration for proxypay
        configs = eval(f"settings.{PP_SETTINGS_CONFIG_KEY}")
        #
        return configs.get(PP_CONFIG_API_TOKEN_KEY)
    #
    return None

# =================================================================================================================================================


from django.contrib import admin
from proxypay.models import Reference
from django.utils.translation import ugettext_lazy as _

class ReferenceAdmin(admin.ModelAdmin):

    # --------------------------------------------------------------------------------------------
	###
	## Settings Properties
	#
    
    list_display = ( 
        'reference',
        'amount', 
        'entity', 
        'payment_local',
        'payment_tarminal',
        'paid_at',
        'created_at',
        'updated_at',
        'expires_in',
        'expired',
        'is_paid', 
    )

    ordering         = ('-created_at', 'updated_at', 'amount')
    search_fields    = ('reference', 'amount')
    list_filter      = ('is_paid', 'entity')

    date_hierarchy   = 'created_at'


    ###
    ## 
    #

    def expired(self, obj):
        return obj.expired

    def payment_local(self, obj):
        if obj.payment:
            return obj.payment.get('terminal_location')
        return None

    def payment_tarminal(self, obj):
        if obj.payment:
            return obj.payment.get('terminal_type')
        return None
            
    expired.boolean = True
    payment_local.short_description = _('Paid In')
    payment_tarminal.short_description = _('Paid With')

    ###
    ## Permissions
    #

    def has_add_permission(self, *args, **kwargs):
        return False

    def has_delete_permission(self, *args, **kwargs):
        return False

    def has_change_permission(self, *args, **kwargs):
        return False

admin.site.register(Reference, ReferenceAdmin)


# Generated by Django 3.1 on 2021-01-08 11:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('proxypay', '0004_reference_paid_at'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='reference',
            name='payment_status',
        ),
        migrations.AddField(
            model_name='reference',
            name='status',
            field=models.CharField(default=0, editable=False, max_length=10),
        ),
        migrations.AlterField(
            model_name='reference',
            name='reference',
            field=models.IntegerField(editable=False, verbose_name='Reference'),
        ),
    ]



###
##  Django Proxypay API 
#

# requests 
import requests

# proxypay stuff
from .conf import get_configurations

# ==========================================================================================================
    
"""Responsible for making all requests to proxypay"""

class ProxypayAPI:

    __headers  = {}     # default api headers
    __url      = ''     # base api url
    __entity   = None   # 
    env        = None

    def __init__(self):

        # proxypay configuration from project settings
        configs = get_configurations()

        # setting the headers
        self.__headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/vnd.proxypay.v2+json',
            'Authorization': f"Token {configs.get('token')}"
        }

        # base url
        self.__url    = configs.get('url')
        # entity
        self.__entity = configs.get('entity')
        # envi
        self.env = configs.get('environment')

    # ==========================================================

    ###
    ##  Property Methods
    # 

    @property
    def entity(self):
        return self.__entity

    # ==========================================================
    
    ###
    ##  Interaction With Proxypay
    # 

    # ------------------------ REFERENCES -----------------------

    # get new reference

    def get_reference_id(self):
        
        """
        Get Generated Reference Id from Proxypay
        Returns reference id (int as string) or False 
        """

        r = self.post('/reference_ids')
        # response status
        return r.json() if r.status_code == 200 else False

    # create or update a payment reference

    def create_or_update_reference(self, reference_id, data):

        """
        Creates or Update a Pament Reference by Reference Id
        Returns Proxypay Payment Reference Object 

        data = { amount, end_datetime, custom_fields }
        """

        r = self.put(f"/references/{reference_id}", data=data)
        # response status
        return True if r.status_code == 204 else False

    # delete reference

    def delete_reference(self, reference_id):

        """
        Delete a reference from Proxypay
        """

        r = self.delete(f"/references/{reference_id}")
        # response status
        return True if r.status_code == 204 else False


    # ------------------------ PAYMENTS -----------------------

    # get unrecognized payments

    def get_payments(self):
        
        """
        Returns a list of all payments that have 
        not yet been recognized
        """

        r = self.get('/payments')
        # response status
        return r.json() if r.status_code == 200 else False

    # check reference payment status

    def check_reference_payment(self, reference_id):

        """
        Checks if a reference has already been paid, if so, 
        returns the payment data and eliminates the payment data in proxyapy

        Make sure to update the reference (proxypay.models.Reference) as paid
        """

        # getting all unrecognized payments
        payments = self.get_payments()

        if payments:
            for payment in payments:
                if payment.get('reference_id') == reference_id:
                    # acknowledge payment
                    self.acknowledge_payment(payment.get('id'))
                    # returns the paymen data
                    return payment
        
        return False

    # Acknowledges payment

    def acknowledge_payment(self, payment_id):

        r = self.delete(f"/payments/{payment_id}")
        # response status
        return True if r.status_code == 204 else False

    # ==========================================================

    ###
    ##  Base Request Methods, GET, POST, PUT, DELETE
    #   

    def get(self, path, params={}):

        """ makes a GET request, path parameter must init with / """

        with requests.get(f"{self.__url}{path}", params=params, headers=self.__headers,) as r:
            return r

    def post(self, path, data={}, params={}):

        """ makes a POST request, path parameter must init with / """

        with requests.post(f"{self.__url}{path}", json=data, params=params, headers=self.__headers,) as r:
            return r

    def put(self, path, data={}, params={}):

        """ makes a PUT request, path parameter must init with / """

        with requests.put(f"{self.__url}{path}", json=data, params=params, headers=self.__headers,) as r:
            return r

    def delete(self, path, data={}, params={}):

        """ makes a DELETE request, path parameter must init with / """

        with requests.delete(f"{self.__url}{path}", json=data, params=params, headers=self.__headers,) as r:
            return r
    
    # ==========================================================
    

# ==========================================================================================================

api = ProxypayAPI()


# Generated by Django 3.0.5 on 2020-05-12 10:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('proxypay', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='reference',
            name='expires_in',
            field=models.DateTimeField(default=None, null=True),
        ),
    ]



# Generated by Django 3.0.5 on 2020-05-11 10:56

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Reference',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('reference', models.IntegerField(editable=False, unique=True, verbose_name='Reference')),
                ('amount', models.DecimalField(decimal_places=2, editable=False, max_digits=12, verbose_name='Amount')),
                ('entity', models.CharField(default=None, editable=False, max_length=100, null=True, verbose_name='Entity')),
                ('custom_fields_text', models.TextField(default='', editable=False)),
                ('payment_status', models.CharField(default='wait', editable=False, max_length=10)),
                ('payment_data_text', models.TextField(default=None, editable=False, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created At')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Update At')),
            ],
        ),
    ]



###
##  Django Proxypay Payment Test
#

# django stuff
from django.core.management.base import BaseCommand

# proxypay stuff
from proxypay.api import api
from proxypay.models import Reference
from proxypay.conf import PP_CONFIG_DEFAULT_DEVELOPMENT_ENV

# =====================================================================================================================

class Command(BaseCommand):

    help = 'For testing payments with Proxypay in the development environment'

    def add_arguments(self, parser):
        parser.add_argument('command', nargs='+', type=str)

    def handle(self, *args, **options):

        # command
        args = options.pop('command')

        # test payment
        if args[0] == 'pay':

            if api.env != PP_CONFIG_DEFAULT_DEVELOPMENT_ENV:
                raise Exception('It is only possible to make fictitious payments in a sandbox environment')

            # getting the reference id from args
            reference_id = int(args[1])
            # get Reference model instance
            reference = Reference.objects.get(reference=reference_id)
            # paying
            r = api.post('/payments', data={
                'amount': float(reference.amount),
                'reference_id': reference.reference
            })
            # r status
            if r.status_code == 200:
                # paid successfully
                payment = r.json()
                #
                reference.paid(payment)
                # 
                self.stdout.write(self.style.SUCCESS(f"Reference: <{reference.reference}>, paid successfully"))
                #
                api.acknowledge_payment(payment.get('id'))
            else:
                self.stdout.write(self.style.ERROR(f"Proxypay returns {r.status_code} status code from API"))


# Generated by Django 3.1 on 2020-10-25 21:10

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('proxypay', '0002_reference_expires_in'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='reference',
            name='custom_fields_text',
        ),
        migrations.RemoveField(
            model_name='reference',
            name='payment_data_text',
        ),
        migrations.AddField(
            model_name='reference',
            name='fields',
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name='reference',
            name='is_paid',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='reference',
            name='payment',
            field=models.JSONField(default=None, null=True),
        ),
    ]



# Generated by Django 3.1 on 2020-10-26 17:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('proxypay', '0003_auto_20201025_2110'),
    ]

    operations = [
        migrations.AddField(
            model_name='reference',
            name='paid_at',
            field=models.DateTimeField(default=None, null=True),
        ),
    ]



# python stuffs
import datetime

# django stuff
from django.utils.timezone import now

from proxypay.exceptions import ProxypayException
from proxypay.conf import get_default_reference_expires_days

# ==========================================================================================================

def get_validated_data(amount, fields={}, days=None):

    ###
    ## Custom Fields
    #

    if len(fields) > 9:
        # 
        raise ProxypayException('Error creating reference, <fields> Add 9 max custom fields')
    
    # reference data
    data         = { 'amount': amount, 'custom_fields': fields }
    end_datetime = None
    
    ###
    ## days to expires
    #

    if days is None:
        # getting days from congurations if set
        days = get_default_reference_expires_days()

    if days and type(days) is int:
        # end_datetime
        end_datetime = now() + datetime.timedelta(days=days)
        # data
        data['end_datetime'] = end_datetime.strftime("%Y-%m-%d")
        #
        data['datetime'] = end_datetime

    elif days:
        # days value error
        raise ProxypayException('Error creating reference, <days> must be an integer')

    return data


###
##  Django Proxypay Payment Reference Creation
#



# proxypay stuffs
from proxypay.api import api
from proxypay.models import Reference
from proxypay.references.utils import get_validated_data
from proxypay.conf import PP_AUTO_PAYMENT_REF_ID

# ==========================================================================================================
 
def create(amount, fields={}, days=None):

    """
    Request to proxypay to create a reference and
    returns an instance of proxypay.models.Reference
    """

    tryTimes = 3

    while tryTimes > 0:
        tryTimes -= 1
        # Get Generated reference id from proxypay
        referenceId = api.get_reference_id()
        if not Reference.objects.is_available(referenceId):
            continue
        #
        data        = get_validated_data(amount, fields, days)
        data['custom_fields'][PP_AUTO_PAYMENT_REF_ID] = str(referenceId)
        datetime    = data.pop('datetime')
        # trying to create the reference
        if api.create_or_update_reference(referenceId, data):
            # saving to the database
            return Reference.objects.create(
                reference=referenceId,
                amount=amount,
                entity=api.entity,
                fields=fields,
                # By default, proxypay references expire at the end of each day
                expires_in=datetime.replace(hour=23,minute=59,second=59)
            )
        else:
            break
    return False



###
##  Django Proxypay Reference Model
#

# django stuff
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.utils.timezone import now
from dateutil import parser
from django.db.models.constraints import UniqueConstraint

# proxypay stuff
from proxypay.api import api
from proxypay.references.utils import get_validated_data
from proxypay.exceptions import ProxypayException
from proxypay.signals import reference_paid, reference_created

# ==========================================================================================================

###
## Paymentt Status
#

PAYMENT_STATUS_WAITING = 0
PAYMENT_STATUS_PAID = 1
PAYMENT_STATUS_CHOICES = (
    (PAYMENT_STATUS_WAITING , 'Waiting'),
    (PAYMENT_STATUS_PAID , 'Paid')
)

# ==========================================================================================================

"""Proxypay References Model Manager"""

class ReferenceModelManager(models.Manager):

    def is_available(self, reference):
        return not self.filter(
            reference=reference,
            status=PAYMENT_STATUS_WAITING,
            expires_in__gt=now()
        ).exists() if reference else False

    def get_reference(self, reference):
        return self.filter(
            reference=reference,
            status=PAYMENT_STATUS_WAITING,
            expires_in__gt=now()
        ).first()

    def create(self, **kwargs):
        # creating the signal
        reference = super(ReferenceModelManager, self).create(**kwargs)
        # Dispatching Signal
        reference_created.send(
            reference.__class__, 
            reference=reference
        )
        #
        return reference

"""Proxypay References Model"""

class Reference(models.Model):

    ###
    ## Model Attributes
    #

    # reference id
    reference           = models.IntegerField(verbose_name=_('Reference'), editable=False)
    amount              = models.DecimalField(verbose_name=_('Amount'), max_digits=12, decimal_places=2, editable=False)
    entity              = models.CharField(verbose_name=_('Entity'), max_length=100, null=True, default=None, editable=False)
    fields              = models.JSONField(default=dict)
    # reference payment status: paid, expired, waiting
    status              = models.CharField(max_length=10, default=PAYMENT_STATUS_WAITING, editable=False)
    payment             = models.JSONField(default=None, null=True)

    is_paid    = models.BooleanField(default=False)
    paid_at    = models.DateTimeField(default=None, null=True)
    # date
    expires_in = models.DateTimeField(null=True, default=None)
    created_at = models.DateTimeField(verbose_name=_('Created At'), auto_now_add=True)
    updated_at = models.DateTimeField(verbose_name=_('Update At'), auto_now=True)


    ###
    ##  Manager
    #

    objects = ReferenceModelManager()

    # --------------------------------------------------------------------------------------------
    ###
    ##  Methods
    #

    def delete(self):

        """Delete payment reference from Proxypay and Database"""

        # deleting from proxypay
        deleted = api.delete_reference(self.reference)
        # 
        if deleted:
            return super(Reference, self).delete()

        raise ProxypayException('Error when trying to delete the reference in the Proxypay')

    def paid(self, payment_data):

        """
        Update reference payment status to paid
        Suitable for use with Proxypay's Webhook
        """

        if not self.payment:
            
            # passando os dados de pagamento na instancia
            self.payment = payment_data
            self.status = PAYMENT_STATUS_PAID
            self.is_paid = True

            try:
                self.paid_at = parser.isoparse(self.payment.get('datetime'))
            except:
                self.paid_at = now()
                
            self.save()
            self.__dispatch_paid_signal()


    def check_payment(self):

        """
        Checks whether the referral payment has already been processed.
        Initially check on the instanse, if it is not processed, check the Proxypay API to be sure. 
        Returns payment data or false
        """

        if not self.payment:
            # check from api
            payment = api.check_reference_payment(self.reference)
            # case already paid
            if payment:
                self.paid(payment)
            # returns payment data or false
            return payment
        # returning the payment data already registered
        return self.payment
    

    def update(self):
        if self.expired:
            data        = get_validated_data(float(self.amount), self.fields)
            datetime    = data.pop('datetime')
            # updating
            if api.create_or_update_reference(self.reference, data=data):
                self.expires_in = datetime.replace(hour=23,minute=59,second=59)
                self.save()
                return True
        return False
            
    # --------------------------------------------------------------------------------------------
    ###
    ##  Property Methods
    #

    @property
    def expired(self):
        if self.expires_in:
            return self.expires_in < now()
        return False

    # --------------------------------------------------------------------------------------------
    ###
    ##  Class Methods
    #

    def __str__(self):
        return f"Proxypay Reference: {self.reference}"

    def __repr__(self):
        return f"Proxypay Reference: {self.reference}"

    # --------------------------------------------------------------------------------------------
    ###
    ##  Signals
    #

    def __dispatch_paid_signal(self):
        # Dispatching Signal
        reference_paid.send(
            self.__class__, 
            reference=self
        )
        


###
##  Django Proxypay Get Reference
#

# proxypay stuffs
from proxypay.models import Reference

# ==========================================================================================================
 
def get (reference_id):

    """
    Get a payment reference instance from proxypay.models.Reference
    Returns false if the passed id reference is not found
    """

    return Reference.objects.get_reference(reference=reference_id)


###
##  Django Proxypay Views
#

# pyhton stuffs
import hmac, hashlib, json
from proxypay.models import reference

# django stuff
from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt

# proxypay stuffs
from proxypay.references import get
from proxypay.conf import get_private_key
from proxypay.conf import PP_AUTO_PAYMENT_REF_ID

# ==============================================================================================

def check_signature(signature, raw_http_body):

    """Check Proxypay Signature"""

    # getting private key from settings.py
    private_key = get_private_key()
    # calc the signature
    calc_signature = hmac.new(
        bytearray(private_key, 'utf-8'),
        msg=raw_http_body,
        digestmod=hashlib.sha256
    ).hexdigest()
    # chack vall
    return signature == calc_signature

# ==============================================================================================

@csrf_exempt
def watch_payments (request):

    """View to watch Proxyapy API Webhook"""

    if request.method == 'POST':
        # check signature
        if check_signature(request.headers.get('X-Signature'), request.body):
            # payment data
            payment = json.loads(request.body)
            # getting reference id
            reference_id = payment.get('custom_fields', {}).get(PP_AUTO_PAYMENT_REF_ID)
            print(payment)
            print(reference_id)
            print(payment.get('reference_id'))
            reference = get(reference_id)
            # chack reference
            if reference:
                # update as paid
                reference.paid(payment)
                # paiment done
                return HttpResponse(status=200)
            # reference not found
            return HttpResponse(status=404)
        # forbidden signatue
        return HttpResponse(status=403)
    # method not allowed
    return HttpResponse(status=405)



#
import os, re, setuptools
from proxypay import __version__ as VERSION

# Long Description
with open("README.md", "r") as readme_file:
    readme = readme_file.read()

setuptools.setup(
    name="django-proxypay",
    version=VERSION,
    description=(
        "Django Proxypay is a Django Framework application/library that facilitates the integration of your Django project with the"
        "Proxypay API. Allowing to generate referrals, recognize payments and look through Proxypay's webhooks"
    ),
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Ageu Matheus",
    author_email="ageumatheus1@gmail.com",
    maintainer="Ageu Matheus",
    maintainer_email="ageumatheus1@gmail.com",
    url="https://github.com/AgeuMatheus/django-proxypay",
    project_urls={
        "Changelog": (
            "https://github.com/AgeuMatheus/django-proxypay"
            + "/blob/master/HISTORY.md"
        )
    },
    packages=setuptools.find_packages(),
    license="MIT License",
    keywords=["django", "payments", "proxypay", "django-proxypay" ],
    install_requires=[
        "Django>=3",
        "requests>=2"
    ],
    python_requires=">=3.6",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)